﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Common;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SilenceORM.DbUtility
{
    public class DbContext<Connection> where Connection : DbConnection, new()
    {
        private string m_ConnString;
        private DbProvider<Connection> m_Provider;

        public DbContext(string connString)
        {
            if (string.IsNullOrWhiteSpace(connString))
                throw new ArgumentNullException();
            m_ConnString = connString;
            m_Provider = new DbProvider<Connection>();
        }

        public DbParameter CreateParameter(string parmName, object value)
        {
            return CreateParameter(parmName, value, new DbType?(), new ParameterDirection?());
        }

        public DbParameter CreateParameter(string parmName, object value, DbType dbType)
        {
            return CreateParameter(parmName, value, new DbType?(dbType), new ParameterDirection?());
        }

        public DbParameter CreateParameter(string parmName, object value, ParameterDirection direction)
        {
            return CreateParameter(parmName, value, new DbType?(), new ParameterDirection?(direction));
        }

        public DbParameter CreateParameter(string parmName, object value, DbType? dbType, ParameterDirection? direction)
        {
            if (string.IsNullOrWhiteSpace(parmName))
                throw new ArgumentNullException(parmName);

            DbParameter parm = m_Provider.CreateParameter();
            parm.ParameterName = parmName;
            parm.Value = value == null ? DBNull.Value : value;
            if (dbType.HasValue)
                parm.DbType = dbType.Value;
            if (direction.HasValue)
                parm.Direction = direction.Value;
            return parm;
        }

        public int ExecuteNonQuery(string cmdText, params DbParameter[] parms)
        {
            return ExecuteNonQuery(cmdText, CommandType.Text, parms);
        }

        public int ExecuteNonQuery(string cmdText, CommandType cmdType, params DbParameter[] parms)
        {
            Func<DbConnection, DbCommand, int> executeSQL = (conn, cmd) => cmd.ExecuteNonQuery();
            return ExecuteDb<int>(cmdText, cmdType, executeSQL, parms);
        }

        public TResult ExecuteScalar<TResult>(string cmdText, params DbParameter[] parms)
        {
            return ExecuteScalar<TResult>(cmdText, CommandType.Text, parms);
        }

        public TResult ExecuteScalar<TResult>(string cmdText, CommandType cmdType, params DbParameter[] parms)
        {
            return ExecuteDb<TResult>(cmdText, cmdType, (conn, cmd) => (TResult)cmd.ExecuteScalar(), parms);
        }

        public DbDataReader ExecuteDataReader(string cmdText, params DbParameter[] parms)
        {
            return ExecuteDataReader(cmdText, CommandType.Text, parms);
        }

        public DbDataReader ExecuteDataReader(string cmdText, CommandType cmdType, params DbParameter[] parms)
        {
            //Connection必须在Reader读取数据后才可关闭，所以单独处理
            DbConnection conn = m_Provider.CreateConnection(m_ConnString);
            using (DbCommand cmd = m_Provider.CreateCommand(conn))
            {
                m_Provider.PrepareCommand(conn, cmd, null, cmdType, cmdText, parms);
                return cmd.ExecuteReader(CommandBehavior.CloseConnection);
            }
        }

        public int ExecuteNonQuery(Func<DbCommand, int> tranExecuteSQL)
        {
            using (DbConnection conn = m_Provider.CreateConnection(m_ConnString))
            {
                DbTransaction tran = m_Provider.CreateTransaction(conn);
                using (DbCommand cmd = m_Provider.CreateCommand(conn))
                {
                    int num;
                    m_Provider.PrepareCommand(conn, cmd, tran);
                    try
                    {
                        num = tranExecuteSQL(cmd);
                        tran.Commit();
                    }
                    catch
                    {
                        tran.Rollback();
                        throw;
                    }
                    return num;
                }
            }
        }

        private TResult ExecuteDb<TResult>(string cmdText, CommandType cmdType, Func<DbConnection, DbCommand, TResult> executeSQL, params DbParameter[] parms)
        {
            using (DbConnection conn = m_Provider.CreateConnection(m_ConnString))
            {
                using (DbCommand cmd = m_Provider.CreateCommand(conn))
                {
                    m_Provider.PrepareCommand(conn, cmd, null, cmdType, cmdText, parms);
                    return executeSQL(conn, cmd);
                }
            }
        }
    }
}
