﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Common;
using System.Threading.Tasks;
using System.Reflection;
using System.Data;
using System.Collections.Concurrent;

using SilenceORM.DbUtility;
using SilenceORM.CommonUtility;
using SilenceORM.TableConfig;

namespace SilenceORM.EF
{
    internal static class EFToSQL<T> where T:new()
    {
        private static ConcurrentDictionary<T, KeyValuePair<TableAttribute, Dictionary<string, ColumnAttribute>>>
            caches = new ConcurrentDictionary<T, KeyValuePair<TableAttribute, Dictionary<string, ColumnAttribute>>>();

        /// <summary>
        /// 生成SQL语句
        /// </summary>
        /// <typeparam name="T">类型</typeparam>
        /// <param name="model">实体</param>
        /// <param name="operate">操作类型</param>
        /// <param name="operateSql">条件语句</param>
        /// <returns>Item1：操作sql语句；Item2：参数数组；Item3：列名-属性名</returns>
        internal static Tuple<string, DbParameter[], Dictionary<string, string>> OperateSQL(T model, Operate operate, string operateSql)
        {
            var modelInfo = GetCache(model);
            Dictionary<string, ColumnAttribute> pmc = GetPropertyMappColumn(modelInfo.Key.Prefix, modelInfo.Value);

            StringBuilder sql = new StringBuilder(100);
            LinkedList<DbParameter> parms = new LinkedList<DbParameter>();

            //操作表语句（表名不可以通过@参数形式传递，所以只能采用拼接）
            sql.AppendFormat(operateSql, modelInfo.Key.TableName);

            if (model != null)
            {
                switch (operate)
                {
                    case Operate.SELECT:
                        SelectSQL(model, pmc, sql, parms);
                        break;
                    case Operate.INSERT:
                        InsertSQL(model, pmc, sql, parms);
                        break;
                    case Operate.UPDATE:
                        UpdateSQL(model, pmc, sql, parms);
                        break;
                    case Operate.DELETE:
                        DeleteSQL(model, pmc, sql, parms);
                        break;
                }
            }

            return new Tuple<string, DbParameter[], Dictionary<string, string>>(sql.ToString(), parms.ToArray(),
                pmc.ToDictionary(c => c.Key, c => c.Value == null ? c.Key : c.Value.ColumnName));

        }

        /// <summary>
        /// 设置列名称
        /// </summary>
        /// <param name="prefix">前缀</param>
        /// <param name="properties">属性信息集合</param>
        /// <returns>Key：属性名；Value：列特性</returns>
        private static Dictionary<string, ColumnAttribute> GetPropertyMappColumn(string prefix, Dictionary<string, ColumnAttribute> properties)
        {
            foreach (string key in properties.Keys)
            {
                properties[key].ColumnName = string.Format("{0}{1}", prefix, properties[key].ColumnName ?? key);
            }

            return properties;
        }

        /// <summary>
        /// 获取主键属性
        /// </summary>
        /// <param name="properties">属性集合</param>
        /// <returns>主键信息（Key：属性名称；Value：列特性）</returns>
        private static Dictionary<string, ColumnAttribute> GetPrimaryKey(Dictionary<string, ColumnAttribute> properties)
        {
            return properties.Where(c => c.Value.IsPrimaryKey).ToDictionary(c => c.Key, c => c.Value);
        }

        private static void SelectSQL(T model, Dictionary<string, ColumnAttribute> pmc, StringBuilder sql, LinkedList<DbParameter> parms)
        {
            var sqlParms = ProvieParameter(model, pmc, (colName, value) => string.Format("{0}=@{1}", colName, value), parms);

            sql.Append("\r\n WHERE ");
            sql.Append(string.Join(" and ", sqlParms.Item2));
        }

        private static void DeleteSQL(T model, Dictionary<string, ColumnAttribute> pmc, StringBuilder sql, LinkedList<DbParameter> parms)
        {
            var sqlParms = ProvieParameter(model, pmc.Where(c => c.Value.IsPrimaryKey).ToDictionary(c => c.Key, c => c.Value), (colName, value) => string.Format("{0}=@{1}", colName, value), parms);

            sql.Append("\r\n WHERE ");
            sql.Append(string.Join(" and ", sqlParms.Item2));
        }

        private static void InsertSQL(T model, Dictionary<string, ColumnAttribute> pmc, StringBuilder sql, LinkedList<DbParameter> parms)
        {
            var sqlParms = ProvieParameter(model, pmc.Where(c => !c.Value.IsIdentity).ToDictionary(c => c.Key, c => c.Value),
                (colName, value) => string.Format("@{0}", value), parms);

            //插入列
            sql.AppendFormat("({0})", string.Join(",", sqlParms.Item1));

            //列值
            sql.AppendFormat("\r\n VALUES({0})", string.Join(",", sqlParms.Item2));
        }

        private static void UpdateSQL(T model, Dictionary<string, ColumnAttribute> pmc, StringBuilder sql, LinkedList<DbParameter> parms)
        {
            var sqlParms = ProvieParameter(model, pmc.Where(c => !(c.Value.IsIdentity || c.Value.IsPrimaryKey)).ToDictionary(c => c.Key, c => c.Value),
                (colName, proName) => string.Format("{0}=@{1}", colName, proName), parms);

            //更新
            sql.AppendFormat(" SET {0}", string.Join(",", sqlParms.Item2));

            var primaryKeys = GetPrimaryKey(pmc);
            var conditionParms = ProvieParameter(model, primaryKeys, (colName, proName) => string.Format("{0}=@{1}", colName, proName), parms);
            sql.AppendFormat(" WHERE {0}", string.Join(" and ", conditionParms.Item2));
        }

        /// <summary>
        /// 整理参数
        /// </summary>
        /// <typeparam name="T">类型</typeparam>
        /// <param name="model">实体类</param>
        /// <param name="pmc">属性-列特性</param>
        /// <param name="condFormat">条件格式</param>
        /// <param name="parms">sql参数</param>
        /// <returns>Item1：有值得属性名称；Item2：条件</returns>
        private static Tuple<LinkedList<string>, LinkedList<string>> ProvieParameter(T model, Dictionary<string, ColumnAttribute> pmc, Func<string, string, string> condFormat, LinkedList<DbParameter> parms)
        {
            //属性集合
            LinkedList<string> propertyNames = new LinkedList<string>();
            //条件集合
            LinkedList<string> conditions = new LinkedList<string>();

            foreach (string key in pmc.Keys)
            {
                //属性值
                object propertyValue = model.GetType().GetProperty(key).GetValue(model);
                if (propertyValue == null)
                    continue;
                propertyNames.AddLast(key);

                //条件
                if (condFormat != null)
                    conditions.AddLast(condFormat(pmc[key].ColumnName, key));

                //参数
                parms.AddLast(SqlContext.GetInstance().CreateParameter(string.Format("@{0}", key), propertyValue));
            }

            return new Tuple<LinkedList<string>, LinkedList<string>>(propertyNames, conditions);
        }

        private static KeyValuePair<TableAttribute, Dictionary<string, ColumnAttribute>> GetCache(T model)
        {
            if (!caches[model].Equals(default(KeyValuePair<TableAttribute, Dictionary<string, ColumnAttribute>>)))
                return caches[model];
            else
            {
                caches[model] = ReflectorHelper.GetPropertiesForModel<T>();
                return caches[model];
            }
        }
    }
}
